import sys
import os
import json
import ttkbootstrap as tb

if sys.version_info[0] >= 3:
    from tkinter import *
    from tkinter import ttk
    from tkinter import font as tkFont
    from tkinter import filedialog as tkFileDialog
    from tkinter import messagebox as tkMessageBox
else:
    from Tkinter import *
    import ttk
    import tkFont
    import tkFileDialog
    import tkMessageBox



from FoxDotEditor.tkimport import *
from renardo_lib.Settings import FONT
from .Format import *
from FoxDotEditor.AppFunctions import index as get_index
from FoxDotEditor.Format import get_keywords
import re


class TextPrompt:
    num_items = 6
    pady = 2

    def __init__(self, root):
        self.root = root
        self.master = self.root.text  # text box
        # ToDo // sort out the name space to check for suggestions
        # keywords = list(self.root.namespace["get_keywords"]())
        keywords = list(get_keywords())
        synthdefs = list(self.root.namespace["SynthDefs"])
        attributes = list(self.root.namespace["Player"].get_attributes())
        player_methods = ["every", "often", "sometimes", "rarely"]
        pattern_methods = list(self.root.namespace["Pattern"].get_methods())
        scales = list(self.root.namespace["Scale"].names())
        other = ["SynthDefs"]
        self.namespace = sorted(list(set(keywords + synthdefs + attributes +
                                         player_methods + pattern_methods +
                                         scales + other)))

        self.values = [StringVar() for n in range(self.num_items)]
        self.clear()

        self.labels = [Label(self.master,
                             textvariable=self.values[n],
                             font=self.root.codefont,
                             foreground="White",
                             anchor=NW,
                             pady=self.pady) for n in range(self.num_items)]

        self.selected = 0
        self.bg = colour_map['prompt_bg']
        self.fg = colour_map['prompt_fg']
        self.suggestions = []
        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.re = re.compile(r"(\w+)$")
        self.__visible = True
        self.hide()

    def get_start_of_word(self):
        """
        Returns the TK index and characters directly preceeding the
        INSERT marker
        """
        # Get contents of this line up to the INSERT
        row, col = get_index(self.master.index(INSERT))
        start, end = "{}.0".format(row), "{}.{}".format(row, col)
        text = self.master.get(start, end)
        match = self.re.search(text)

        if match is not None:
            index, text = "{}.{}".format(row, match.start()), match.group(0)
            self.anchor = match.span()
        else:
            index, text = "1.0", ""
            self.anchor = None
        return (index, text)

    def in_word(self):
        """ Returns True if the next character is alphanumeric """
        return self.master.get(INSERT).isalnum()

    def show(self):
        """ Displays the prompt with suggestions """
        if not self.__visible:
            return
        # 1. Get location - start of the word
        index, word = self.get_start_of_word()

        # If there is a alphanumeric character next, dont show
        if len(word) == 0 or self.in_word():
            return self.hide()
        bbox = self.master.bbox(index)

        if bbox is not None:
            self.selected = 0
            self.x, self.y, self.w, self.h = bbox
            # 3. Find first 4 words
            self.suggestions = self.find_suggestions(word)
            # If there is only 1 suggestion and we're at the end of the word
            # (or no suggestions), just hide
            num_suggestions = len(self.suggestions)

            if num_suggestions == 0 or (num_suggestions == 1 and (word == self.suggestions[0])):
                return self.hide()
            # 4. Show
            self.update_values(self.suggestions)
            self.move(self.x, self.y)
            self.visible = True
        return

    def move(self, x, y):
        offset = self.h
        width = max((len(val.get()) for val in self.values))
        for i, label in enumerate(self.labels):
            if self.values[i].get() != "":
                label.place(x=x, y=y + offset)
                label.config(width=width,
                             bg=(self.fg if self.selected == i else self.bg),)
                offset += (self.h + (self.pady * 2))
            else:
                label.place(x=9999, y=9999)
        return

    def hide(self):
        self.clear()
        self.move(x=9999, y=9999)
        self.visible = False
        return

    def clear(self):
        for value in self.values:
            value.set("")
        return

    def update_values(self, values):
        self.clear()
        for i, word in enumerate(values[:self.num_items]):
            self.values[i].set(word)
        return

    def find_suggestions(self, word):
        words = []
        i = 0
        for phrase in self.namespace:
            # if phrase.lower().startswith(word.lower()):
            if phrase.startswith(word):
                words.append(phrase)
                i += 1
            if i == self.num_items:
                break
        return words

    def cycle_up(self):
        self.selected = max(0, self.selected - 1)
        return self.move(self.x, self.y)

    def cycle_down(self):
        self.selected = min(len(self.suggestions) - 1, self.selected + 1)
        return self.move(self.x, self.y)

    def autocomplete(self):
        """ Inserts the remainder of the currently highlited suggestion """
        if self.anchor is not None:
            LENGTH = self.anchor[1] - self.anchor[0]
            self.master.delete("{}-{}c".format(INSERT, LENGTH), INSERT)
            WORD = self.values[self.selected].get()
            self.master.insert(INSERT, WORD)
            self.root.update()
        self.hide()
        return

    def toggle(self):
        """ Flags the prompt to show / not show automatically """
        self.__visible = not self.__visible
        return



from FoxDotEditor.tkimport import *
from FoxDotEditor.Format import *
from FoxDotEditor.AppFunctions import *
from renardo_lib.Settings import AUTO_COMPLETE_BRACKETS

whitespace = [" ", "\t", "\n", "\r", "\f", "\v"]


class BracketHandler:
    counter = 0

    def __init__(self, master):
        self.root = master
        self.text = master.text
        self.active = AUTO_COMPLETE_BRACKETS
        self.inbrackets = False
        self.separators = py_separators
        self.style = bracket_formatting
        self.left_brackets = dict(zip(left_b, right_b))
        self.right_brackets = dict(zip(right_b, left_b))
        self.right_brackets_only = list("]})")
        self.queue = []
        if self.active:
            for char in list(self.left_brackets.keys()) + list(self.right_brackets.keys()):
                # this binds bracket presses to the handle method
                self.text.bind(char, self.handle)
        self.visible_range = (0, 0)

    def handle(self, event=None, insert=INSERT):
        self.visible_range = self.text.get_visible_range()
        ret = None
        line, column = index(self.text.index(insert))
        # 1. Type a left bracket
        next_char = self.text.get(index(line, column))
        if event.char in self.left_brackets:
            # A. Type a left bracket and an area is selected
            try:
                a = self.text.index(SEL_FIRST)
                b = self.text.index(SEL_LAST)
                self.text.insert(b, self.left_brackets[event.char])
                self.text.insert(a, event.char)
                # unselect
                self.text.tag_remove(SEL, "1.0", END)
                ret = "break"
            except Exception:
                pass
            # B. If next character is a whitespace or a right bracket add a
            # closing bracket as well.
            if next_char in whitespace + self.right_brackets_only:
                # If the next char is quote mark, just move the cursor
                if event.char in ("'", '"'):
                    if event.char == next_char:
                        self.text.mark_set(insert, index(line, column + 1))
                        ret = "break"
                    else:
                        # Only add a closing quote mark if there are already
                        # an even no. of quotes in the line
                        text = self.text.get(index(line, 0),
                                             index(line, "end"))
                        n = text.count(event.char)
                        if n % 2 == 0:
                            self.text.insert(self.text.index(insert),
                                             event.char + self.left_brackets[event.char])
                            self.text.mark_set(insert, index(line, column + 1))
                            ret = "break"
                else:
                    self.text.insert(self.text.index(insert),
                                     event.char + self.left_brackets[event.char])
                    self.text.mark_set(insert, index(line, column + 1))
                    ret = "break"
            # Update line colours
            self.root.colour_line(line)
        # 2. Type right bracket
        elif event.char in self.right_brackets:
            # Go back and find an open bracket -> assume the first open bracket
            # is related to *this* one
            self.text.insert("{}.{}".format(line, column),
                             event.char)  # Add bracket
            coords = self.find_starting_bracket(line, column, event.char)
            new_line, new_col = line, column + 1
            # Assume we are adding a new bracket
            adding_bracket = True
            # Get index of the end of the buffer
            end_line, end_col = index(self.text.index("{}.end".format(self.visible_range[1])))
            while (new_line, new_col) != (end_line, end_col):
                # If we find a closing bracket,  find it's pair
                next_char = self.text.get(index(new_line, new_col))
                if next_char == event.char:
                    coords_ = self.find_starting_bracket(new_line,
                                                         new_col,
                                                         event.char,
                                                         offset=0)
                    # If there is a closing brackets
                    if coords_ is None:
                        adding_bracket = False
                        break
                    else:
                        adding_bracket = True
                if index(new_line, new_col) == self.text.index(index(new_line,
                                                                     "end")):
                    new_line += 1
                    new_col = 0
                else:
                    new_col += 1

            if not adding_bracket:
                loc = index(line, column+1)
                if self.text.get(loc) == event.char:
                    self.text.delete(loc)
            ret = "break"
            # Update line colours
            self.root.colour_line(line)
            # Highlight brackets (remove old highlighting first)
            self.text.tag_delete("tag_open_brackets")

            if coords is not None:
                row, col = coords
                self.text.tag_config("tag_open_brackets", **self.style)
                self.text.tag_add("tag_open_brackets", "{}.{}".format(row, col),
                                  "{}.{}".format(row, col+1))
                self.text.tag_add("tag_open_brackets", "{}.{}".format(line, column),
                                  "{}.{}".format(line, column+1))
        if ret is None:
            self.text.insert(INSERT, event.char)
            self.root.colour_line(line)
        # Store the text
        self.root.text_as_string = self.root.get_all()
        # Add separator to undo stack
        self.text.edit_separator()
        return "break"

    def delete(self, insert=INSERT):
        """
        Deletes enclosed set of brackets if no contents and returns True,
        returns False all other times
        """
        if self.active:
            line, column = index(self.text.index(insert))
            next_char = self.text.get(index(line, column))
            prev_char = self.text.get(index(line, column-1))
            if prev_char in self.left_brackets:
                if next_char == self.left_brackets[prev_char] and column > 0:
                    self.text.delete(index(line, column-1), index(line, column+1))
                    return True
        return False

    def find_starting_bracket(self, line, column, bracket_style, offset=0):
        """
        Finds the opening bracket to the closing bracket at line, column
        co-ords. Returns None if not found. Only goes as far as the first line
        visible on screen
        """
        line_length = column - 1
        used_br = offset
        start_row, end_row = self.visible_range  # returns tuple
        for row in range(line, start_row - 1, -1):
            for col in range(line_length, -1, -1):
                # If the char is a left bracket and not used, break
                if self.text.get("{}.{}".format(row, col)) == self.right_brackets[bracket_style]:
                    if used_br == 0:
                        return row, col
                    else:
                        used_br -= 1
                elif self.text.get("{}.{}".format(row, col)) == bracket_style:
                    used_br += 1
            line_length = int(self.text.index("{}.end".format(row-1)).split(".")[1])
        else:
            return None



#!/usr/bin/python

"""

FoxDot is a Python library and programming environment that provides a fast and
user-friendly abstraction to the powerful audio-engine, SuperCollider. It comes
with its own IDE, which means it can be used straight out of the box; all you
need is Python and SuperCollider and you're ready to go!

Copyright Ryan Kirkbride 2015
"""

import sys
from renardo_lib.runtime import *


def boot_supercollider():
    """ Uses subprocesses to boot supercollider from the cli """
    import time
    import platform
    import os
    import subprocess
    import getpass
    try:
        import psutil
    except ImportError:
        os.system("pip install psutil")
        import sys
        sys.exit("Installed psutil, please start FoxDot again.")

    sclangpath = ""  # find path to sclang
    thispath = ""  # find this path
    thisdir = os.getcwd()
    OS = platform.system()
    username = getpass.getuser()

    if (OS == "Windows"):
        sclangloc = os.popen('where /R "C:\\Program Files" sclang.exe').read()
        thiscwd = str(sclangloc)
        ourcwd = thiscwd.replace('\\sclang.exe\n', '')

        def is_proc_running(name):
            for p in psutil.process_iter(attrs=["name", "exe", "cmdline"]):
                # print(p);
                procname = p.info['name'] or \
                     p.info['exe'] and os.path.basename(p.info['exe']) == name\
                     or p.info['cmdline'] and p.info['cmdline'][0] == name
                if (procname.startswith(name)):
                    return True
            return False

        running = (is_proc_running("sclang"))

        if (running is False):
            startup = thisdir+"/FoxDot/startup.scd"
            # os.system("sclang"+startup+" &")
            subprocess.Popen([sclangloc, startup], cwd=ourcwd, shell=True)

    elif (OS == "Linux"):
        def is_proc_running(name):
            for p in psutil.process_iter(attrs=["name", "cmdline"]):
                # print(p);
                procname = p.info['name'] or \
                     p.info['cmdline'] and p.info['cmdline'][0] == name
                if (procname.startswith(name)):
                    return True

        running = (is_proc_running("sclang"))

        if (running is False):
            startup = thisdir+"/FoxDot/startup.scd"
            # os.system('sclang "/home/foxdot/Desktop/FoxDot-Cross-Platform/
            # FoxDot/startup.scd" &') #functional
            os.system("sclang "+startup+" &")

    else:
        print("Operating system unrecognised")
        # Potentially get the user to choose their OS from a list?
        # Then run the corresponding functions#


if "--boot" in sys.argv:
    boot_supercollider()
    sys.argv.remove("--boot")


def main():
    """ Function for starting the GUI when importing the library """
    from FoxDotEditor.Editor import workspace
    FoxDot = workspace(FoxDotCode).run()


def Go():
    """ Function to be called at the end of Python files with FoxDot code in to
    keep the TempoClock thread alive. """
    try:
        import time
        while 1:
            time.sleep(100)
    except KeyboardInterrupt:
        return
